/*
 * Copyright 2007 The Apache Software Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License")
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package net.sf.hibernate4gwt.core;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import javax.persistence.EntityManagerFactory;

import net.sf.hibernate4gwt.core.beanlib.IClassMapper;
import net.sf.hibernate4gwt.core.beanlib.proxy.ProxyClassMapper;
import net.sf.hibernate4gwt.core.store.IPojoStore;
import net.sf.hibernate4gwt.core.store.stateless.StatelessPojoStore;
import net.sf.hibernate4gwt.exception.CloneException;
import net.sf.hibernate4gwt.exception.PojoStoreException;
import net.sf.hibernate4gwt.exception.TransientHibernateObjectException;
import net.sf.hibernate4gwt.pojo.ILazyPojo;
import net.sf.hibernate4gwt.rebind.ProxyClassLoader;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.hibernate.SessionFactory;

/**
 * Manager for Hibernate POJO handling
 * @author bruno.marchesson
 *
 */
public class HibernateBeanManager
{
	//----
	// Attributes
	//----
	/**
	 * The unique instance of the Hibernate Helper
	 */
	private static HibernateBeanManager _instance = null;
	
	/**
	 * Log channel
	 */
	private Log _log = LogFactory.getLog(HibernateBeanManager.class);
	
	/**
	 * The associated POJO Handler
	 */
	private IPojoStore _pojoStore;
	
	/**
	 * The Class mapper
	 */
	private IClassMapper _classMapper;
		
	/**
	 * The POJO lazy killer
	 */
	private LazyKiller _lazyKiller;
	
	//----
	// Property
	//----
	/**
	 * @return the unique instance of the singleton
	 */
	public synchronized static HibernateBeanManager getInstance()
	{
		if (_instance == null)
		{
			_instance = new HibernateBeanManager();
		}
		return _instance;
	}
	
	/**
	 * @return the pojo store
	 */
	public IPojoStore getPojoStore()
	{
		return _pojoStore;
	}
	
	/**
	 * set the used pojo store
	 */
	public void setPojoStore(IPojoStore pojoStore)
	{
		_log.info("Using PojoStore : " + pojoStore);
		_pojoStore = pojoStore;
	}
	
	/**
	 * @return the class mapper
	 */
	public IClassMapper getClassMapper()
	{
		return _classMapper;
	}
	
	/**
	 * @param mapper the class Mapper to set
	 */
	public void setClassMapper(IClassMapper mapper)
	{
		_log.info("Using class mapper : " + mapper);
		_classMapper = mapper;
		_lazyKiller.setClassMapper(mapper);
	}

	/**
	 * Helper setter to initialize HibernateHelper 
	 * Must be called on startup !
	 */
	public void setSessionFactory(SessionFactory sessionFactory)
	{
		HibernateHelper.getInstance().setSessionFactory(sessionFactory);
	}
	
	/**
	 * Helper setter to initialize HibernateHelper with the embedded 
	 * Hibernate SessionFactory.
	 * Will not work with other JPA entity manager
	 * Lust be called on startup !
	 */
	public void setEntityManagerFactory(EntityManagerFactory entityManagerFactory)
	{
		setSessionFactory(((org.hibernate.ejb.HibernateEntityManagerFactory) entityManagerFactory).getSessionFactory()); 
	}
	
	//-------------------------------------------------------------------------
	//
	// Constructor
	//
	//-------------------------------------------------------------------------
	/**
	 * Empty Constructor
	 */
	protected HibernateBeanManager()
	{
		_pojoStore = new StatelessPojoStore();
		_lazyKiller = new LazyKiller();
		
		// default class mapper
		setClassMapper(new ProxyClassMapper());
	}
		
	//-------------------------------------------------------------------------
	//
	// Hibernate Java 1.4 POJO public interface
	//
	//-------------------------------------------------------------------------
	/**
	 * Clone and store the Hibernate POJO(s)
	 */
	public Object clone(Object object)
	{
	//	Precondition checking
	//
		if (object == null)
		{
			return null;
		}
	
	//	Collection handling
	//
		if (object instanceof Collection)
		{
			return cloneCollection((Collection)object);
		}
		else if (object instanceof Map)
		{
			return cloneMap((Map)object);
		}
		else
		{
			return clonePojo(object);
		}
	}
	
	/**
	 * Clone and store the Hibernate POJO(s)
	 */
	public Object merge(Object object)
	{
	//	Precondition checking
	//
		if (object == null)
		{
			return null;
		}
	
	//	Collection handling
	//
		if (object instanceof Collection)
		{
			return mergeCollection((Collection)object);
		}
		else if (object instanceof Map)
		{
			return mergeMap((Map)object);
		}
		else
		{
			return mergePojo(object);
		}
	}
	
	/**
	 * Remove the object of the pojo store
	 */
	public void remove(Object object)
	{
	//	Precondition checking
	//
		if (object == null)
		{
			return;
		}
	
	//	Collection handling
	//
		if (object instanceof Collection)
		{
			removeCollection((Collection)object);
		}
		else if (object instanceof Map)
		{
			removeMap((Map)object);
		}
		else
		{
			removePojo(object);
		}
	}
	
	//-------------------------------------------------------------------------
	//
	// Hibernate Java 1.4 POJO internal methods
	//
	//-------------------------------------------------------------------------
	/**
	 * Clone and store the Hibernate POJO
	 */
	protected Object clonePojo(Object pojo)
	{
	//	Precondition checking
	//
		if (HibernateHelper.getInstance().isHibernatePojo(pojo) == false)
		{
			return pojo;
		}
		
	//	Store the pojo
	//
		_pojoStore.store(pojo);
		
	//	Clone the pojo
	//
		return _lazyKiller.detach(pojo);
	}
	
	/**
	 * Clone and store a map of Hibernate POJO
	 */
	protected Map cloneMap(Map hibernatePojoMap)
	{
	//	Clone each element of the map
	//
		Map cloneMap = new HashMap();
		
		Iterator iterator = hibernatePojoMap.entrySet().iterator();
		while (iterator.hasNext())
		{
			Entry entry = (Entry) iterator.next();
			cloneMap.put(clone(entry.getKey()), 
						 clone(entry.getValue()));
		}
		
		return cloneMap;
	}
	
	/**
	 * Clone and store a collection of Hibernate POJO
	 */
	protected Collection cloneCollection(Collection hibernatePojoList)
	{
	//	Clone each element of the collection
	//
		Collection clonePojoList = createNewCollection(hibernatePojoList);
		for (Object hibernatePojo : hibernatePojoList)
		{
			clonePojoList.add(clone(hibernatePojo));
		}
		
		return clonePojoList;
	}
	
	/**
	 * Retrieve the Hibernate Pojo and merge the modification from GWT 
	 * @param clonePojo the clone pojo
	 * @return the merged Hibernate POJO
	 * @exception UnsupportedOperationException if the clone POJO does not 
	 * implements ILazyPojo and the POJO store is stateless
	 */
	protected Object mergePojo(Object clonePojo)
	{
	//	Get Hibernate associated class
	//
		Class hibernateClass = _classMapper.getSourceClass(clonePojo.getClass());
		if (hibernateClass == null)
		{
		//	Not a clone : take the inner class
		//
			hibernateClass = clonePojo.getClass();
		}
		
	//	Precondition checking
	//
		if (HibernateHelper.getInstance().isHibernateClass(hibernateClass) == false)
		{
		//	Not an Hibernate class : no merge needed
		//
			return clonePojo;
		}
		
		if ((clonePojo instanceof ILazyPojo == false) &&
			(_pojoStore.isStateless() == true))
		{
			throw new UnsupportedOperationException("Cannot merge Pure POJO without a stateful POJO store !\n"+
													clonePojo.getClass().getName() + " should implement ILazyPojo...");
		}
		
	//	Retrieve the pojo
	//
		Object hibernatePojo = null;
		try
		{
			hibernatePojo = _pojoStore.restore(clonePojo, hibernateClass);
			if (hibernatePojo == null)
			{
				throw new PojoStoreException("Hibernate POJO not found !", clonePojo);
			}
		}
		catch(TransientHibernateObjectException ex)
		{
			_log.info("Transient object : no merge needed  _ " + clonePojo);
			if (hibernateClass.equals(clonePojo.getClass()) == true)
			{
			//	Not a proxy : we can return the clone pojo
			//
				return clonePojo;
			}
			else
			{
			//	Proxy : generate a new Hibernate pojo and merge the proxy with it
			//
				try
				{
					hibernatePojo = hibernateClass.newInstance();
				}
				catch(Exception e)
				{
					throw new RuntimeException("Cannot create a fresh new instance of the class " + hibernateClass, e);
				} 
			}
		}
		
	//	Merge the modification in the Hibernate Pojo
	//
		_lazyKiller.attach(hibernatePojo, clonePojo);
		return hibernatePojo;
	}
	
	/**
	 * Retrieve the Hibernate Pojo List and merge the modification from GWT 
	 * @param clonePojoList the clone pojo list
	 * @return a list of merged Hibernate POJO 
	 * @exception UnsupportedOperationException if a POJO from the list does 
	 * not implements ILazyPojo and the POJO store is stateless
	 */
	protected Collection mergeCollection(Collection clonePojoList)
	{
		Collection hibernatePojoList = createNewCollection(clonePojoList);
		
	//	Retrieve every hibernate from pojo list
	//
		for (Object clonePojo : clonePojoList)
		{
			try
			{
				hibernatePojoList.add(merge(clonePojo));
			}
			catch(TransientHibernateObjectException e)
			{
			//	Keep new pojo (probably created from GWT)
			//
				hibernatePojoList.add(clonePojo);
			}
		}
		
		return hibernatePojoList;
	}
	
	/**
	 * Fill copy map with Hibernate merged POJO
	 * @param cloneMap
	 * @return a map with merge Hibernate POJO
	 */
	protected Map mergeMap(Map cloneMap)
	{
		Map hibernateMap = new HashMap();
		
	//	Iterate over map
	//
		Iterator iterator = cloneMap.entrySet().iterator();
		while (iterator.hasNext())
		{
			Entry entry = (Entry) iterator.next();
			
			// Merge key
			Object key = entry.getKey();
			try
			{
				key = merge(key);
			}
			catch (TransientHibernateObjectException ex)
			{ /* keep key untouched */ }
			
			// Merge value
			Object value = entry.getValue();
			try
			{
				value = merge(value);
			}
			catch (TransientHibernateObjectException ex)
			{ /* keep value untouched */ }
			
			hibernateMap.put(key, value);
		}
		
		return hibernateMap;
	}
	
	/**
	 * Remove the pojo from the pojo store
	 * @param pojo hibernate pojo
	 */
	protected void removePojo(Object pojo)
	{
		_pojoStore.remove(pojo);
	}
	
	/**
	 * Remove the content of the map of pojo from the pojo store
	 * @param pojoList hibernate pojo list
	 */
	protected void removeMap(Map pojoMap)
	{
		Iterator iterator = pojoMap.entrySet().iterator();
		while (iterator.hasNext())
		{
			Entry entry = (Entry) iterator.next();
			
			remove(entry.getKey());
			remove(entry.getValue());
		}
	}
	/**
	 * Remove the collection of pojo from the pojo store
	 * @param pojoList hibernate pojo collection
	 */
	protected void removeCollection(Collection pojoList)
	{
		for (Object pojo : pojoList)
		{
			remove(pojo);
		}
	}
	
	/**
	 * Create a new collection with the same behavior than the argument one
	 * @param pojoCollection the source collection
	 * @return a newly created, empty collection
	 */
	protected Collection createNewCollection(Collection pojoCollection)
	{
		if (pojoCollection instanceof List)
		{
			return new ArrayList(pojoCollection.size());
		}
		else if (pojoCollection instanceof Set)
		{
			return new HashSet(pojoCollection.size());
		}
		else
		{
			throw new CloneException("Unhandled collection type : " + pojoCollection.getClass().toString());
		}
	}
}
