package net.sf.hibernate4gwt.core;

import java.io.Serializable;
import java.lang.reflect.Method;

import net.sf.hibernate4gwt.exception.NotHibernateObjectException;
import net.sf.hibernate4gwt.exception.TransientHibernateObjectException;

import org.hibernate.EntityMode;
import org.hibernate.SessionFactory;
import org.hibernate.metadata.ClassMetadata;

/**
 * Helper singleton for Hibernate access
 * Centralize the SessionFactory and add some usefull methods
 * @author BMARCHESSON
 */
public class HibernateHelper
{
	//----
	// Attributes
	//----
	/**
	 * The unique instance of the singleton
	 */
	private static HibernateHelper _instance = null;
	
	/**
	 * The Hibernate session factory
	 */
	private SessionFactory _sessionFactory;	

	//----
	// Properties
	//----
	/**
	 * @return the unique instance of the singleton
	 */
	public static HibernateHelper getInstance()
	{
		if (_instance == null)
		{
			_instance = new HibernateHelper();
		}
		return _instance;
	}
	
	/**
	 * @return the hibernate session Factory
	 */
	public SessionFactory getSessionFactory()
	{
		return _sessionFactory;
	}

	/**
	 * @param sessionFactory the factory to set
	 */
	public void setSessionFactory(SessionFactory sessionFactory)
	{
		_sessionFactory = sessionFactory;
	}

	//-------------------------------------------------------------------------
	//
	// Constructor
	//
	//-------------------------------------------------------------------------
	/**
	 * Empty constructor
	 */
	protected HibernateHelper()
	{
	}
	
	//-------------------------------------------------------------------------
	//
	// Public interface
	//
	//-------------------------------------------------------------------------
	/**
	 * @return the ID of the argument Hibernate Pojo
	 */
	public Serializable getId(Object pojo)
	{
		return getId(pojo, pojo.getClass());
	}
	
	/**
	 * @return the ID of the argument DTO with the same name than the Hibernate class
	 */
	public Serializable getId(Object pojo, Class hibernateClass)
	{
	//	Precondition checking
	//
		if (_sessionFactory == null)
		{
			throw new NullPointerException("No Hibernate Session Factory defined !");
		}
		
	//	Retrieve Class hibernate metadata
	//
		ClassMetadata hibernateMetadata = _sessionFactory.getClassMetadata(hibernateClass);
		if (hibernateMetadata == null)
		{
		//	Not an hibernate Class !
		//
			return new NotHibernateObjectException(hibernateClass);
		}
		
	//	Retrieve ID
	//
		Serializable id = null;
		
		if (hibernateClass.equals(pojo.getClass()))
		{
		//	the pojo has the same class, simple use metadata
		//
			id = hibernateMetadata.getIdentifier(pojo, EntityMode.POJO);	
		}
		else
		{
		//	DTO case : invoke the method with the same name
		//
			String property = hibernateMetadata.getIdentifierPropertyName();
			
			try
			{
				// compute getter method name
				property = property.substring(0,1).toUpperCase() + 
						   property.substring(1);
				String getter = "get" + property;
				
				// Find getter method
				Class pojoClass = pojo.getClass();
				Method method = pojoClass.getMethod(getter, (Class[])null);
				if (method == null)
				{
					throw new RuntimeException("Cannot find method " + getter + " for class " + pojoClass);
				}
				id = (Serializable) method.invoke(pojo,(Object[]) null);
			}
			catch (Exception ex)
			{
				throw new RuntimeException("Invocation exception ", ex);
			}
		}
		
	//	Post condition checking
	//
		if (id == null)
		{
			throw new TransientHibernateObjectException(pojo);
		}
		return id;
	}
	
	/**
	 * Indicates if the pojo is managed by Hibernate of not
	 */
	public boolean isHibernatePojo(Object pojo)
	{
		if (pojo == null)
		{
			return false;
		}
		return isHibernateClass(pojo.getClass());
	}
	
	/**
	 * Indicates if the class is managed by Hibernate of not
	 */
	public boolean isHibernateClass(Class clazz)
	{
	//	Precondition checking
	//
		if (_sessionFactory == null)
		{
			throw new NullPointerException("No Hibernate Session Factory defined !");
		}
		
	//	Retrieve ID of the POJO
	//
		return (_sessionFactory.getClassMetadata(clazz) != null);
	}
}
