/*
 * Copyright 2007 The Apache Software Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License")
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package net.sf.hibernate4gwt.core;

import net.sf.beanlib.hibernate.HibernateBeanReplicator;
import net.sf.beanlib.hibernate3.Hibernate3BeanReplicator;
import net.sf.beanlib.hibernate3.Hibernate3CollectionReplicator;
import net.sf.beanlib.provider.BeanPopulator;
import net.sf.beanlib.provider.collector.PrivateSetterMethodCollector;
import net.sf.beanlib.spi.BeanTransformerSpi;
import net.sf.beanlib.spi.CustomBeanTransformerSpi;
import net.sf.hibernate4gwt.core.beanlib.CloneBeanReplicator;
import net.sf.hibernate4gwt.core.beanlib.CloneMapper;
import net.sf.hibernate4gwt.core.beanlib.FromHibernateToLazyBeanPopulatable;
import net.sf.hibernate4gwt.core.beanlib.FromLazyBeanToHibernatePopulatable;
import net.sf.hibernate4gwt.core.beanlib.IClassMapper;
import net.sf.hibernate4gwt.core.beanlib.MappedClassBeanReplicator;
import net.sf.hibernate4gwt.core.beanlib.TimestampCustomTransformer;
import net.sf.hibernate4gwt.rebind.ProxyManager;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * This class replaces all "lazy but not loaded" Hibernate association with null
 * to allow the argument POJO to be used in other libraries without additional loading
 * of 'LazyInitializationException'.  
 * The lazy properties are stored as string. They can be reused for "reattaching" the clone
 * pojo to a fresh Hibernate object.
 * @author bruno.marchesson
 */
public class LazyKiller
{
	//----
	// Attributes
	//----
	/**
	 * Log channel
	 */
	private static Log _log = LogFactory.getLog(LazyKiller.class);
	
	/**
	 * The clone class mapper
	 */
	private CloneMapper _cloneMapper;

	/**
	 * The merge class mapper
	 */
	private CloneMapper _mergeMapper;
	
	//----
	// Properties
	//----
	/**
	 * @param mapper the class Mapper to set
	 */
	public void setClassMapper(IClassMapper mapper)
	{
	//	Create clone and merge wrapping mappers
	//
		_cloneMapper = new CloneMapper(mapper, true);
		_mergeMapper = new CloneMapper(mapper, false);
	}
	
	//-------------------------------------------------------------------------
	//
	// Constructor
	//
	//-------------------------------------------------------------------------
	/**
	 * Empty constructor
	 */
	public LazyKiller()
	{
		this(null);
	}
	
	/**
	 * Base constructor
	 * @param cloneMapper the clone class mapping
	 */
	public LazyKiller(IClassMapper classMapper)
	{
		setClassMapper(classMapper);
	}

	//------------------------------------------------------------------------
	//
	// Public interface
	//
	//------------------------------------------------------------------------
	/**
	 * Hibernate detachment
	 * @param hibernatePojo the input hibernate pojo
	 * @return a pure Java clone
	 */
	public Object detach(Object hibernatePojo)
	{
	//	Precondition checking
	//
		if (hibernatePojo == null)
		{
			return null;
		}
			
		_log.info("Detaching " + hibernatePojo.toString());
		
	// 	Search for Proxy
	//
		Class proxyClass = ProxyManager.getInstance().getProxyClass(hibernatePojo.getClass());
		if (proxyClass != null)
		{
			// Generate proxy
			return clone(hibernatePojo, proxyClass);
		}
		else
		{
			// No proxy mode
			return clone(hibernatePojo, hibernatePojo.getClass());
		}
	}
	
	/**
	 * Hibernate detachment
	 * @param hibernatePojo the input hibernate pojo
	 * @return a pure Java clone
	 */
	public Object detach(Object hibernatePojo, Class cloneClass)
	{
	//	Precondition checking
	//
		if (hibernatePojo == null)
		{
			return null;
		}
			
		_log.info("Detaching " + hibernatePojo.toString());
		
	//	 Clone with beanLib
	//
		return clone(hibernatePojo, cloneClass);
	}
	
	/**
	 * Hibernate attachment
	 * @param hibernatePojo the stored or fresh Hibernate POJO
	 * @param clonePojo the cloned pojo
	 */
	public void attach(Object hibernatePojo, Object clonePojo)
	{
    //	Precondition checking
	//
		if ((hibernatePojo == null) ||
			(clonePojo == null))
		{
			return;
		}
		
		_log.info("Attaching " + clonePojo.toString());
		
	//	Populate with BeanLib
	//
		populate(hibernatePojo, clonePojo);
	}
	
	//-------------------------------------------------------------------------
	//
	// Internal methods
	//
	//-------------------------------------------------------------------------
	/**
	 * Clone the abstract POJO with BeanLib
	 * Every time a lazy property is detected, it is replaced with null.
	 * It is also marked as "lazy" for ILazyPojo sub-classes
	 * @param pojo
	 * @return
	 */
	protected Object clone(Object hibernatePojo, Class cloneClass)
	{
		HibernateBeanReplicator replicator = null;
		if (_cloneMapper == null)
		{
		//	No specific clone mapping : use default BeanLib replicator
		//
			replicator = new Hibernate3BeanReplicator();
		}
		else
		{
		//	Java5 support : need class mapper for clone and merge
		//
			replicator = new CloneBeanReplicator(_cloneMapper);
		}
		
	//	Timestamp handling
	//
		replicator.initCustomTransformer(new CustomBeanTransformerSpi.Factory()
		{
			public CustomBeanTransformerSpi newCustomBeanTransformer(final BeanTransformerSpi beanTransformer)
			{
				return new TimestampCustomTransformer(beanTransformer);
			}
		});
		
	//	Lazy properties handling
	//
		replicator.initDetailedBeanPopulatable(new FromHibernateToLazyBeanPopulatable());
		replicator.initSetterMethodCollector(PrivateSetterMethodCollector.inst);
		
		return replicator.copy(hibernatePojo, cloneClass);
	}
	
	/**
	 * Populate the hibernatePojo (a fresh new one or the one used to clone)
	 * with the clone detached object.
	 * This Hibernate POJO holds the lazy properties information
	 */
	public void populate(Object hibernatePojo, Object clonePojo)
	{
	//	Populate hibernate POJO from the cloned pojo
	//
		BeanPopulator replicator = BeanPopulator.newBeanPopulator(clonePojo, hibernatePojo);
		
		if (_mergeMapper != null)
		{
		//	Java5 support : need class mapper for clone and merge
		//
			BeanTransformerSpi transfomer = (BeanTransformerSpi) replicator.getTransformer();
			transfomer.initBeanReplicatable(MappedClassBeanReplicator.factory);
			
			MappedClassBeanReplicator beanReplicator = (MappedClassBeanReplicator)
														transfomer.getBeanReplicatable();
			beanReplicator.setCloneMapper(_mergeMapper);
		}
		else
		{
		//	Hibernate collection replicatable
		//
			BeanTransformerSpi transfomer = (BeanTransformerSpi) replicator.getTransformer();
			transfomer.initCollectionReplicatable(Hibernate3CollectionReplicator.getFactory());
		}
		
	//	Timestamp handling
	//
		((BeanTransformerSpi) replicator.getTransformer()).initCustomTransformer(new CustomBeanTransformerSpi.Factory()
		{
			public CustomBeanTransformerSpi newCustomBeanTransformer(final BeanTransformerSpi beanTransformer)
			{
				return new TimestampCustomTransformer(beanTransformer);
			}
		});
		
	//	Lazy properties handling
	//
		replicator.initDetailedBeanPopulatable(new FromLazyBeanToHibernatePopulatable());
		replicator.initSetterMethodCollector(PrivateSetterMethodCollector.inst);
		
		replicator.populate();
	}
}
