/*
 * Copyright 2007 The Apache Software Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License")
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.sf.hibernate4gwt.core.beanlib;

import java.lang.reflect.Method;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.hibernate.Hibernate;

import net.sf.beanlib.spi.DetailedBeanPopulatable;
import net.sf.hibernate4gwt.exception.CloneException;
import net.sf.hibernate4gwt.pojo.ILazyPojo;

/**
 * Populatable for POJO with the lazy information of the Hibernate POJO
 * This populatable is used to fill back an Hibernate POJO from a pure clone
 * POJO (not implementing ILazyPojo).
 * Only the not lazy properties of the Hibernate POJO are filled.
 * To be correct, the clone and merge operations must be managed with the
 * <b>same</b> Hibernate POJO, or a POJO with the <b>same</b> fetching strategy
 * @author bruno.marchesson
 */
public class FromLazyBeanToHibernatePopulatable implements DetailedBeanPopulatable 
{
	//----
	// Properties
	//----
	/**
	 * Log channel
	 */
	private static Log _log = LogFactory.getLog(FromLazyBeanToHibernatePopulatable.class);
	
	//-------------------------------------------------------------------------
	//
	// DetailedBeanPopulatable implementation
	//
	//-------------------------------------------------------------------------
	/* (non-Javadoc)
	 * @see net.sf.beanlib.spi.DetailedBeanPopulatable#shouldPopulate(java.lang.String, java.lang.Object, java.lang.reflect.Method, java.lang.Object, java.lang.reflect.Method)
	 */
	public boolean shouldPopulate(String propertyName, 
								  Object fromBean, 
								  Method readerMethod, 
								  Object toBean, 
								  Method setterMethod)
	{
	//	Is the property lazy loaded ?
	//
		try
		{
			if (fromBean instanceof ILazyPojo == true)
			{
			//	ILazyPojo : just check the lazy properties string
			//
				if (((ILazyPojo) fromBean).isLazyProperty(propertyName) == true)
				{
					_log.info(fromBean.toString()+ "." + propertyName + " --> lazy loaded ");
					return false;
				}
				else
				{
					return true;
				}
			}
			else if (toBean.getClass() != fromBean.getClass())
			{
			//  Cannot use the Hibernate bean with a DTO class (Java5 DTO class) !
			//
				throw new CloneException("Explicit DTO (Java5) must be declared as extending LazyGwtPojo !");
			}
			else
			{
			//	Call the readerMethod on the Hibernate Bean to determine the lazy property
			//
				Object result = readerMethod.invoke(toBean, new Object[] {});
				if (Hibernate.isInitialized(result) == false)
				{
					_log.info(fromBean.toString()+ "." + propertyName + " --> lazy loaded ");
					return false;
				}
				else
				{
					return true;
				}
			}
		}
		catch(Exception ex)
		{
			throw new RuntimeException(ex);
		}
	}

}
